from beep.structure.base import BEEPDatapath
from beep.structure.arbin import ArbinDatapath
from beep.structure.indigo import IndigoDatapath
from beep.structure.maccor import MaccorDatapath
from beep.structure.neware import NewareDatapath
from beep.structure.biologic import BiologicDatapath
from beep.structure.cli import process_file_list_from_json, auto_load, auto_load_processed
